from ruamel.yaml import YAML, yaml_object
from ruamel.yaml.comments import CommentedMap

yaml = YAML()


class AttrYamlMixin:
    @classmethod
    def from_yaml(cls, constructor, node):
        state = CommentedMap()
        constructor.construct_mapping(node, state)
        return cls(**state)



import itertools
import tempfile
from pathlib import Path

import pytest
from WDL import parse_document

from aclimatise import explore_command
from aclimatise.converter.cwl import CwlGenerator
from aclimatise.converter.wdl import WdlGenerator
from aclimatise.model import CliArgument, Flag, SimpleFlagArg
from aclimatise.yaml import yaml

from .util import convert_validate, skip_not_installed

# Note: the parse and explore tests run conversion tests already. These tests are for specific edge cases


def test_premade_samtools(samtools_cmd):
    """
    Use a command tree that was generated beforehand, to quickly detect issues relating to the conversion of command
    trees
    """
    convert_validate(samtools_cmd, explore=True)


def test_premade_bedtools(bedtools_cmd):
    """
    Use a command tree that was generated beforehand, to quickly detect issues relating to the conversion of command
    trees
    """
    convert_validate(bedtools_cmd, explore=True)


@skip_not_installed("samtools")
@skip_not_installed("samtools.pl")
def test_explore_samtools_pl(yaml_converter):
    """
    Tests that commands with a non-standard file extension include their extension in the final output, and don't
    override another command with the same stem
    """
    samtools = explore_command(["samtools"], max_depth=0)
    samtools_pl = explore_command(["samtools.pl"], max_depth=0)
    with tempfile.TemporaryDirectory() as temp_dir:
        path = Path(temp_dir)
        filenames = set()
        for path, command in itertools.chain(
            yaml_converter.generate_tree(samtools, temp_dir),
            yaml_converter.generate_tree(samtools_pl, temp_dir),
        ):
            filenames.add(path.name)

        assert filenames == {"samtools.yml", "samtools.pl.yml"}


def test_docker_conversion(bedtools_cmd):
    intersect = bedtools_cmd["intersect"]
    container = "quay.io/biocontainers/bedtools:2.29.2--hc088bd4_0"
    intersect.docker_image = container
    with tempfile.NamedTemporaryFile() as cwl_file:
        CwlGenerator().save_to_file(intersect, path=Path(cwl_file.name))
        cwl_file.seek(0)
        parsed_cwl = yaml.load(cwl_file)
        assert any(
            [
                hint["class"] == "DockerRequirement" and hint["dockerPull"] == container
                for hint in parsed_cwl["hints"]
            ]
        )

    wdl = WdlGenerator().save_to_string(intersect)
    parsed_wdl = parse_document(wdl).tasks[0]
    assert parsed_wdl.runtime["docker"].literal.value == container


@pytest.mark.parametrize(
    "flag,cwltype,wdltype",
    [
        [
            Flag(
                synonyms=["--some-flag"],
                optional=True,
                args=SimpleFlagArg("string"),
                description="",
            ),
            "string?",
            "String?",
        ],
        [
            Flag(
                synonyms=["--some-flag"],
                optional=False,
                args=SimpleFlagArg("string"),
                description="",
            ),
            "string",
            "String",
        ],
    ],
)
def test_types_conversion(flag: CliArgument, cwltype: str, wdltype: str):
    """
    Test that types are being correctly translated from aCLImatise types to CWL and WDL
    """
    assert CwlGenerator.arg_to_cwl_type(flag) == cwltype
    assert (
        WdlGenerator.type_to_wdl(flag.get_type(), optional=flag.optional).get_string()
        == wdltype
    )



import typing

import attr

from aclimatise.flag_parser.parser import CliParser
from aclimatise.model import Command, Flag
from aclimatise.usage_parser.parser import UsageParser


def parse_help(cmd: typing.Collection[str], text: str, max_length=1000) -> Command:
    """
    Parse a string of help text into a Command. Use this if you already have run the executable and extracted the
    help text yourself

    :param cmd: List of arguments used to generate this help text, e.g. ['bwa', 'mem']
    :param text: The help text to parse
    :param max_length: If the input text has more than this many lines, no attempt will be made to parse the file (as
        it's too large, will likely take a long time, and there's probably an underlying problem if this has happened).
        In this case, an empty Command will be returned
    """
    if len(text.splitlines()) > max_length:
        return Command(list(cmd))

    help_command = CliParser().parse_command(name=cmd, cmd=text)
    usage_command = UsageParser().parse_usage(list(cmd), text)

    # Combine the two commands by picking from the help_command where possible, otherwise falling back on the usage
    fields = dict(
        help_text=text,
        # Use the help command's positionals preferentially, but fall back to usage
        positional=help_command.positional or usage_command.positional,
        # Combine the flags from both help and usage
        named=list(Flag.combine([help_command.named, usage_command.named])),
    )
    for field in attr.fields(Command):
        fields[field.name] = (
            fields.get(field.name)
            or getattr(help_command, field.name)
            or getattr(usage_command, field.name)
        )

    return Command(**fields)



from typing import List

import janis_core as janis
from aclimatise import cli_types
from aclimatise.cli_types import CliType
from aclimatise.converter import NamedArgument, WrapperGenerator
from aclimatise.model import CliArgument, Command, Flag, Positional


class JanisGenerator(WrapperGenerator):
    @classmethod
    def format(cls) -> str:
        return "janis"

    def save_to_string(self, cmd: Command) -> str:

        clt = self.command_to_tool(cmd)
        return clt.translate("janis", to_console=False)

    def command_to_tool(self, cmd: Command) -> janis.CommandToolBuilder:

        inputs: List[CliArgument] = [*cmd.named] + (
            [] if self.ignore_positionals else [*cmd.positional]
        )
        names = self.choose_variable_names(inputs)

        tool = janis.CommandToolBuilder(
            tool=cmd.as_filename,
            base_command=list(cmd.command),
            inputs=self.get_inputs(names),
            outputs=self.get_outputs(names),
            version="v0.1.0",
            container=cmd.docker_image,
        )

        return tool

    def type_to_janis_type(
        self, typ: cli_types.CliType, optional: bool
    ) -> janis.DataType:

        if isinstance(typ, cli_types.CliFile):
            return janis.File(optional=optional)
        elif isinstance(typ, cli_types.CliDir):
            return janis.Directory(optional=optional)
        elif isinstance(typ, cli_types.CliString):
            return janis.String(optional=optional)
        elif isinstance(typ, cli_types.CliFloat):
            return janis.Float(optional=optional)
        elif isinstance(typ, cli_types.CliInteger):
            return janis.Int(optional=optional)
        elif isinstance(typ, cli_types.CliBoolean):
            return janis.Boolean(optional=optional)
        elif isinstance(typ, cli_types.CliEnum):
            return janis.String(optional=optional)
        elif isinstance(typ, cli_types.CliList):
            # TODO: how is Array<String?> represented?
            inner = self.type_to_janis_type(typ.value, optional=False)
            return janis.Array(inner, optional=optional)

        elif isinstance(typ, cli_types.CliTuple):
            return self.type_to_janis_type(
                CliType.lowest_common_type(typ.values), optional=False
            )
        else:
            raise Exception(f"Invalid type {typ}!")

    def arg_to_janis_type(self, arg: CliArgument) -> janis.DataType:
        return self.type_to_janis_type(arg.get_type(), arg.optional)

    def get_inputs(self, names: List[NamedArgument]) -> List[janis.ToolInput]:
        ret = []
        for arg in names:
            assert arg.name != "", arg
            ret.append(
                janis.ToolInput(
                    tag="in_" + arg.name,
                    input_type=self.arg_to_janis_type(arg.arg),
                    position=arg.arg.position
                    if isinstance(arg.arg, Positional)
                    else None,
                    prefix=arg.arg.longest_synonym
                    if isinstance(arg.arg, Flag)
                    else None,
                    doc=arg.arg.description,
                )
            )
        return ret

    def get_outputs(self, names: List[NamedArgument]) -> List[janis.ToolOutput]:
        ret = []
        for arg in names:
            typ = arg.arg.get_type()
            if isinstance(typ, cli_types.CliFileSystemType) and typ.output:
                ret.append(
                    janis.ToolOutput(
                        tag="out_" + arg.name,
                        output_type=self.arg_to_janis_type(arg.arg),
                        doc=arg.arg.description,
                        selector=janis.InputSelector("in_" + arg.name),
                    )
                )
        return ret

    @property
    def suffix(self) -> str:
        return ".py"



import shutil
import tempfile
import traceback
from pathlib import Path

import pyparsing
import pytest
from click.testing import CliRunner
from packaging import version

from aclimatise.cli import main
from aclimatise.yaml import yaml

from .util import skip_not_installed, validate_cwl, validate_janis, validate_wdl


@pytest.fixture()
def runner():
    return CliRunner()


def cli_worked(result):
    if result.exit_code == 0:
        return True
    else:
        traceback.print_exception(*result.exc_info)
        assert result.exit_code == 0


def test_pipe_wdl(runner, htseq_help):
    result = runner.invoke(
        main, ["pipe", "htseq-count", "--format", "wdl"], input=htseq_help
    )
    cli_worked(result)
    validate_wdl(result.output)


def test_pipe_cwl(runner, htseq_help):
    result = runner.invoke(
        main, ["pipe", "htseq-count", "--format", "cwl"], input=htseq_help
    )
    cli_worked(result)
    validate_cwl(result.output)


def test_pipe_janis(runner, htseq_help):
    result = runner.invoke(
        main, ["pipe", "htseq-count", "--format", "janis"], input=htseq_help
    )
    cli_worked(result)
    validate_janis(result.output)


@skip_not_installed("htseq-count")
def test_explore_htseq(runner, caplog):
    caplog.set_level(100000)
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(main, ["explore", "htseq-count", "--out-dir", tempdir])
        cli_worked(result)
        assert len(list(Path(tempdir).iterdir())) == 3


@skip_not_installed("ls")
@skip_not_installed("man")
def test_explore_ls_man(runner, caplog):
    caplog.set_level(100000)
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(main, ["explore", "ls", "--man", "--out-dir", tempdir])
        cli_worked(result)
        with (Path(tempdir) / "ls.yml").open() as fp:
            parsed = yaml.load(fp)
            assert parsed.help_text.startswith("LS(1)")


@skip_not_installed("samtools")
def test_explore_samtools(runner, caplog):
    caplog.set_level(100000)
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            main, ["explore", "samtools", "--help-flag", "--help", "--out-dir", tempdir]
        )
        cli_worked(result)
        assert len(list(Path(tempdir).iterdir())) > 20


@skip_not_installed("samtools")
def test_explore_samtools_no_subcommands(runner, caplog):
    caplog.set_level(100000)
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            main,
            [
                "explore",
                "samtools",
                "--no-subcommands",
                "--out-dir",
                tempdir,
            ],
        )
        cli_worked(result)
        # Since we aren't looking at subcommands, there should be one file for each format
        assert len(list(Path(tempdir).iterdir())) >= 3


@pytest.mark.skipif(
    version.parse(pyparsing.__version__) < version.parse("3.0.0a2"),
    reason="PyParsing 3.0.0+ is not installed",
)
def test_grammar(runner):
    result = runner.invoke(main, ["grammar"])
    assert result.exit_code == 0
    assert len(result.output) > 20



import os
import subprocess
from typing import Collection, List, Optional

from aclimatise.execution import Executor
from aclimatise.integration import parse_help
from aclimatise.model import Command


class ManPageExecutor(Executor):
    def __init__(
        self,
        man_paths: List[str] = [],
        subcommand_sep: Collection[str] = ("-", "_"),
        man_flags: Collection[str] = ["--no-subpages"],
        **kwargs
    ):
        """
        :param man_paths: Additional paths within which to look for man pages
        :param subcommand_sep: A list of separators to use to generate man paths from subcommands. For example
            ``git branch`` has an associated man page at ``git-branch``, using the hyphen as a separator.
        :param man_flags: Additional flags to pass to the ``man`` command
        """
        super().__init__(**kwargs)
        self.man_paths = man_paths
        self.subcommand_sep = subcommand_sep
        self.man_flags = man_flags

    def execute_with_sep(self, command: List[str], separator: str = "-") -> str:
        """
        Returns the man page text for the provided command, using the provided subcommand separator, or an empty string
        if this man page doesn't exist
        """
        env = {**os.environ.copy(), "MANPAGER": "cat"}  # Don't use a pager
        if len(self.man_paths) > 0:
            env.update({"MANPATH": ":".join(self.man_paths)})

        sub_man = separator.join(command)
        result = subprocess.run(
            ["man", *self.man_flags, sub_man],
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if result.returncode == 0:
            return result.stdout.decode()

        return ""

    def convert(self, command: List[str]) -> Command:
        if len(command) == 1:
            return parse_help(
                command, self.execute_with_sep(command), max_length=self.max_length
            )
        else:
            commands = []
            for sep in self.subcommand_sep:
                man_text = self.execute_with_sep(command, sep)
                commands.append(
                    parse_help(command, man_text, max_length=self.max_length)
                )
            return Command.best(commands)



"""
Contains the CLI data model
"""
import abc
import enum
import itertools
import re
import typing
from abc import abstractmethod
from itertools import chain
from operator import attrgetter

import attr
from ruamel.yaml import yaml_object
from word2number import w2n

import aclimatise
from aclimatise import cli_types
from aclimatise.cli_types import CliFileSystemType, CliString
from aclimatise.name_generation import segment_string
from aclimatise.nlp import wordsegment
from aclimatise.usage_parser.model import UsageInstance
from aclimatise.yaml import AttrYamlMixin, yaml


def first(lst: typing.List, default):
    if len(lst) == 0:
        return default
    return lst[0]


def useless_name(name: typing.List[str]):
    """
    Returns true if this name (sequence of strings) shouldn't be used as a variable name because it's too short and
    uninformative. This includes an entirely numeric name, which is almost never what you want
    """
    joined = "".join(name)
    if len(name) < 1 or len(joined) <= 1 or joined.isnumeric():
        return True

    # Numeric names are not useful
    try:
        if all([w2n.word_to_num(tok) is not None for tok in name]):
            return True
    except Exception:
        pass

    return False


@yaml_object(yaml)
@attr.s(
    auto_attribs=True,
)
class Command(AttrYamlMixin):
    """
    Class representing an entire command or subcommand, e.g. `bwa mem` or `grep`
    """

    def __attrs_post_init__(self):
        # Store certain special flags in their own fields
        if self.help_flag is None:
            for flag in self.named:
                if (
                    "--help" in flag.synonyms
                    or "-help" in flag.synonyms
                    or ("-h" in flag.synonyms and isinstance(flag.args, EmptyFlagArg))
                ):
                    self.help_flag = flag
                    self.named.remove(flag)

        if self.version_flag is None:
            for flag in self.named:
                if "--version" in flag.synonyms:
                    self.version_flag = flag
                    self.named.remove(flag)

        if self.usage_flag is None:
            for flag in self.named:
                if "--usage" in flag.synonyms:
                    self.usage_flag = flag
                    self.named.remove(flag)

    def __getitem__(self, item: str) -> "Command":
        """
        If present, returns a subcommand with the following name. For example, samtools_cmd['sort'] will return the
        "samtools sort" command
        """
        for sub in self.subcommands:
            if sub.command == [*self.command, item]:
                return sub
        raise KeyError(
            "{} does not have a subcommand {}".format(" ".join(self.command), item)
        )

    def reanalyse(self, parent: "Command" = None) -> "Command":
        """
        Re-analyses the entire command tree using the existing help text but the current parser, and returns the new tree
        """
        if len(self.subcommands) > 0:
            return Command(
                generated_using=self.generated_using,
                help_text=self.help_text,
                command=self.command,
                subcommands=[cmd.reanalyse(self) for cmd in self.subcommands],
                parent=parent,
            )
        else:
            replacement = aclimatise.parse_help(cmd=self.command, text=self.help_text)
            replacement.parent = parent
            replacement.help_text = self.help_text
            replacement.generated_using = self.generated_using
            return replacement

    def valid_subcommand(self, compare_depth=1) -> bool:
        """
        Returns true if command is a valid subcommand, relative to its parent
        """
        parent = self.ancestor(compare_depth)

        # Recursively call this on all ancestors
        if parent.parent is not None and not self.valid_subcommand(compare_depth + 1):
            return False

        # This isn't a subcommand if it has the same input text as the parent
        if self.help_text and self.help_text == parent.help_text:
            return False

        # This isn't a subcommand if it has no flags
        if len(self.positional) + len(self.named) == 0:
            return False

        # This isn't a subcommand if it shares any positional with the parent command
        for pos_a, pos_b in zip(parent.positional, self.positional):
            if pos_a == pos_b:
                return False

        # This isn't a subcommand if it shares any flags with the parent command
        for flag_a, flag_b in zip(parent.named, self.named):
            if flag_a == flag_b:
                return False

        return True

    @property
    def outputs(self) -> typing.List["CliArgument"]:
        """
        Returns a list of inputs which are also outputs, for example the "-o" flag is normally an input (you have to
        provide a filename), but also an output (you are interested in the contents of the file at that path after the
        command has been run)
        """
        ret = []
        for arg in itertools.chain(self.named, self.positional):
            typ = arg.get_type()
            # Yes this is a bit awkward, the output field should probably be on every type, defaulting to False
            if isinstance(typ, cli_types.CliFileSystemType) and typ.output:
                ret.append(arg)
        return ret

    @property
    def as_filename(self) -> str:
        """
        Returns a sample filename that might be used to store this command (without a suffix)
        """
        return "_".join([token for token in self.command]).replace("-", "_")

    @property
    def empty(self) -> bool:
        """
        True if we think this command failed in parsing, ie it has no arguments
        """
        return (len(self.positional) + len(self.named) + len(self.subcommands)) == 0

    @property
    def depth(self) -> int:
        """
        Returns the "depth" of this command, aka how many ancestors it has. An orphan command has depth 0, a subcommand
        has depth 1, a sub-sub-command has depth 2 etc
        """
        cmd = self
        depth = 0
        while cmd.parent is not None:
            cmd = cmd.parent
            depth += 1

        return depth

    def ancestor(self, levels: int = 1) -> typing.Optional["Command"]:
        """
        Returns the ancestor of the command `levels` steps up the family tree. For example, `cmd.ancestor(1)` is
        equivalent to `cmd.parent`. `cmd.ancestor(2)` is equivalent to `cmd.parent.parent`.
        """
        cmd = self
        for i in range(levels):
            cmd = cmd.parent
            if cmd is None:
                break

        return cmd

    @staticmethod
    def best(commands: typing.Collection["Command"]) -> "Command":
        """
        Given a list of commands generated from the same executable, but with different flags, determine the best one
        and return it
        :param commands:
        :return:
        """
        # Currently we just return the command with the most flags
        return max(
            commands,
            key=lambda com: (
                len(com.named) + len(com.positional),
                # len(com.help_text) if com.help_text else 0,
            ),
        )

    @property
    def all_synonyms(self) -> typing.Set[str]:
        """
        Returns all flag synonyms for all flags in this command
        """
        return set(chain.from_iterable([flag.synonyms for flag in self.named]))

    def command_tree(self) -> typing.Generator["Command", None, None]:
        """
        Returns a generator over the entire command tree. e.g. if this command has 2 subcommands, each with 2
            subcommands, this will return a generator with 7 Commands
        """
        yield self
        for command in self.subcommands:
            yield from command.command_tree()

    command: typing.List[str]
    """
    The command line used to invoke this command, e.g. ["bwa", "mem"]
    """

    positional: typing.List["Positional"] = attr.ib(factory=list)
    """
    All positional arguments supported by this command
    """

    named: typing.List["Flag"] = attr.ib(factory=list)
    """
    All named arguments (flags) supported by this command
    """

    parent: typing.Optional["Command"] = None
    """
    The parent command, if this is a subcommand
    """

    subcommands: typing.List["Command"] = attr.ib(factory=list)
    """
    A list of subcommands of this command, e.g. "bwa" has the subcommand "bwa mem"
    """

    usage: typing.List["UsageInstance"] = attr.ib(factory=list)
    """
    Different usage examples provided by the help
    """

    help_flag: typing.Optional["Flag"] = None
    """
    If identified, this is the flag that returns help text
    """

    usage_flag: typing.Optional["Flag"] = None
    """
    If identified, this is the flag that returns usage examples
    """

    version_flag: typing.Optional["Flag"] = None
    """
    If identified, this is the flag that returns the version of the executable
    """

    help_text: typing.Optional[str] = None
    """
    Optionally, the entire help text that was used to generate this Command
    """

    generated_using: typing.Optional[str] = None
    """
    Optionally, the flag that was used to generate this command. Often this will be the same as the help_flag
    """

    docker_image: typing.Optional[str] = None
    """
    If available, a docker image in which to run this command
    """


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class CliArgument(AttrYamlMixin):
    """
    A generic parent class for both named and positional CLI arguments
    """

    description: str
    """
    Description of the function of this argument
    """

    optional: bool = False
    """
    If true, this argument is not required
    """

    @abstractmethod
    def argument_name(self) -> typing.List[str]:
        return []

    @abstractmethod
    def full_name(self) -> str:
        """
        Return a human-readable representation of this argument
        """
        pass

    @abstractmethod
    def get_type(self) -> cli_types.CliType:
        """
        Return a type object indicating the type of data this argument holds. e.g. If it's an array type this will be a
        CliList.
        """
        pass


@yaml_object(yaml)
@attr.s(auto_attribs=True, kw_only=True)
class Positional(CliArgument):
    """
    A positional command-line argument. This probably means that it is required, and has no arguments like flags do
    """

    def full_name(self) -> str:
        """
        Getting the full name for a positional argument is easy - it's just the parameter name
        """
        return self.name

    position: int
    """
    The position in the command line that this argument must occupy
    """

    name: str
    """
    The name of this argument
    """

    description: str
    """
    A description of the function of this argument
    """

    @staticmethod
    def deduplicate(
        positionals: typing.Collection["Positional"],
    ) -> typing.List["Positional"]:
        key = attrgetter("name")
        sort = sorted(positionals, key=key)
        groups = itertools.groupby(sort, key=key)
        ret = []
        for key, group in groups:
            grp = list(group)
            ret.append(Positional.merge(grp))
        return sorted(ret, key=attrgetter("position"))

    @staticmethod
    def merge(other: typing.List["Positional"]) -> "Positional":
        """
        Combine the information contained within two Positionals, and return the new Positional
        """
        return Positional(
            position=first(
                [pos.position for pos in other if pos.position is not None], -1
            ),
            name=first(
                [
                    pos.name
                    for pos in other
                    if pos.name is not None and len(pos.name) > 0
                ],
                None,
            ),
            # We want this to be optional if it is considered optional in any situation
            optional=any([pos.optional for pos in other]),
            description=max([pos.description for pos in other], key=len, default=None),
        )

    def get_type(self) -> cli_types.CliType:
        # Try the the flag name, then the description in that order

        name_type = infer_type(self.name)
        if name_type is not None:
            return name_type

        return infer_type(self.full_name()) or cli_types.CliString()


@yaml_object(yaml)
@attr.s(auto_attribs=True, kw_only=True)
class Flag(CliArgument):
    """
    Represents one single flag, with all synonyms for it, and all arguments, e.g. `-h, --help`
    """

    optional: bool = True
    """
    If true, this argument is not required
    """

    synonyms: typing.List[str]
    """
    A list of different ways to invoke this same option, e.g. ``-v`` and ``--verbose``
    """

    description: typing.Optional[str]
    """
    A description of the function of this flag
    """

    args: "FlagArg"
    """
    Describes the arguments to this flag, e.g. ``-n 1`` has a single numeric argument
    """

    def __attrs_post_init__(self):
        if self.optional is None:
            # Flags are optional by default
            self.optional = True

    @staticmethod
    def deduplicate(flags: typing.Collection["Flag"]) -> typing.List["Flag"]:
        todo = list(flags[:])
        clusters = []

        while len(todo) > 0:
            # Choose a flag, then find anything it intersects with
            flag_i = todo.pop()
            current_cluster = [flag_i]
            current_synonyms = set(flag_i.synonyms)
            to_remove = []

            for j in range(len(todo)):
                flag_j = todo[j]
                if set(flag_j.synonyms).intersection(current_synonyms):
                    to_remove.append(j)
                    current_synonyms.update(flag_j.synonyms)
                    current_cluster.append(flag_j)

            clusters.append(current_cluster)

        return [Flag.merge(cluster) for cluster in clusters]

    @staticmethod
    def merge(flags: typing.List["Flag"]) -> "Flag":
        """
        Combine the information contained within two Flags, and return the new Flag
        """
        return Flag(
            # Take the longest description
            description=max([flag.description for flag in flags], key=len),
            # Just union the two lists
            synonyms=list(set().union(*[flag.synonyms for flag in flags])),
            args=first([flag.args for flag in flags if flag.args], EmptyFlagArg()),
            optional=any([flag.optional for flag in flags]),
        )

    @staticmethod
    def combine(
        flag_lists: typing.Iterable[typing.Iterable["Flag"]],
    ) -> typing.Iterable["Flag"]:
        """
        Combines the flags from several sources, choosing the first one preferentially
        """
        lookup = {}

        # Build a list of flags, but only ever choose the first instance of a synonym
        for flags in flag_lists:
            for flag in flags:
                for synonym in flag.synonyms:
                    stripped = synonym.lstrip("-")
                    if stripped not in lookup:
                        lookup[stripped] = flag

        # Now, make them unique by description
        unique = {flag.longest_synonym: flag for flag in lookup.values()}

        return list(unique.values())

    def argument_text(self) -> typing.List[str]:
        return self.args.text()

    def variable_name(
        self, description_name: typing.List[str] = []
    ) -> typing.List[str]:
        """
        Returns a list of words that should be used in a variable name for this argument
        """
        # The super method returns the best name from the flag name or the description. If neither is sufficient, use
        # the argument to generate a name
        best = super().variable_name(description_name)

        if useless_name(best):
            nfa = list(self._name_from_arg)
            if not useless_name(nfa):
                return nfa

        return best

    @property
    def _name_from_arg(self) -> typing.Iterable[str]:
        """
        Generate a 1-3 word variable name for this flag, by parsing the description
        """
        if self.args is not None and hasattr(self.args, "name"):
            return segment_string(self.args.name)
        else:
            return []

    def get_type(self) -> cli_types.CliType:
        # Try the argument name, then the flag name, then the description in that order
        arg_type = self.args.get_type()
        flag_type = infer_type(self.full_name())
        desc_type = infer_type(self.description)

        candidates = [
            cand
            for cand in [
                arg_type,
                flag_type,
                desc_type,
            ]
            if cand is not None
        ]
        if len(candidates) == 0:
            return CliString()

        # TODO: make this sorting a bit better using CliType.representable
        prioritised = sorted(
            candidates,
            key=lambda typ: (
                not isinstance(typ, CliString),  # First, prioritise non-strings,
                isinstance(typ, CliFileSystemType)
                and typ.output,  # Next, prioritise output types
            ),
            reverse=True,
        )

        return prioritised[0]

    def full_name(self) -> str:
        """
        Getting the full name for a named flag is slightly harder, we need to find the longest synonym
        """
        return self.longest_synonym

    @staticmethod
    def from_synonyms(
        synonyms: typing.Iterable["FlagSynonym"], description: typing.Optional[str]
    ):
        """
        Creates a usable Flag object by combining the synonyms provided
        """
        synonym_str = []
        args = None
        arg_count = float("-inf")

        for synonym in synonyms:
            synonym_str.append(synonym.name)
            if synonym.argtype.num_args() > arg_count:
                arg_count = synonym.argtype.num_args()
                args = synonym.argtype

        return Flag(synonyms=synonym_str, args=args, description=description)

    @property
    def longest_synonym(self) -> str:
        """
        Returns the longest synonym this flag has. e.g. for `-h, --help`, it will return `--help`
        """
        return max(self.synonyms, key=lambda synonym: len(synonym))

    @property
    def shortest_synonym(self) -> str:
        """
        Returns the shortest synonym this flag has. e.g. for `-h, --help`, it will return `-h`
        """
        return min(self.synonyms, key=lambda synonym: len(synonym))


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class FlagSynonym(AttrYamlMixin):
    """
    Internal class for storing the arguments for a single synonym
    """

    name: str
    """
    The entire flag string, e.g. "-n" or "--lines"
    """

    argtype: "FlagArg"
    """
    The number and type of arguments that this flag takes
    """

    @property
    def capital(self):
        return "".join(
            [
                segment.capitalize()
                for segment in re.split("[-_]", self.name.lstrip("-"))
            ]
        )


int_re = re.compile(
    r"\b((int(eger)?)|size|length|max|min|(num(ber)?))\b", flags=re.IGNORECASE
)
str_re = re.compile(r"\bstr(ing)?\b", flags=re.IGNORECASE)
float_re = re.compile(r"\b(float|decimal)\b", flags=re.IGNORECASE)
bool_re = re.compile(r"\bbool(ean)?\b", flags=re.IGNORECASE)
file_re = re.compile(r"\b(file(name|path)?|path)\b", flags=re.IGNORECASE)
input_re = re.compile(r"input", flags=re.IGNORECASE)
output_re = re.compile(r"\bout(put)?\b", flags=re.IGNORECASE)
dir_re = re.compile(r"\b(folder|directory)\b", flags=re.IGNORECASE)

float_num_re = re.compile(
    r"[+-]?(([0-9]*\.[0-9]+)|((?:0|[1-9]\d*)(?:\.\d*)?(?:[eE][+\-]?\d+)))",
    flags=re.IGNORECASE,
)
int_num_re = re.compile(r"([+-]?[0-9]+)", flags=re.IGNORECASE)


def distinguish_inout(
    string, cls: typing.Type[cli_types.CliFileSystemType]
) -> cli_types.CliFileSystemType:
    """
    distinguish input/output files/directories given a string
    """
    im = input_re.search(string)
    om = output_re.search(string)
    if not im and om:
        return cls(output=True)
    else:
        return cls(output=False)


def infer_type(string) -> typing.Optional[cli_types.CliType]:
    """
    Reads a string (argument description etc) to find hints about what type this argument might be. This is
    generally called by the get_type() methods
    """
    if bool_re.search(string):
        return cli_types.CliBoolean()
    elif float_re.search(string):
        return cli_types.CliFloat()
    elif int_re.search(string):
        return cli_types.CliInteger()
    elif file_re.search(string):
        return distinguish_inout(string, cli_types.CliFile)
    elif dir_re.search(string):
        return distinguish_inout(string, cli_types.CliDir)
    elif str_re.search(string):
        return cli_types.CliString()
    elif float_num_re.search(string):
        return cli_types.CliFloat()
    elif int_num_re.search(string):
        return cli_types.CliInteger()
    else:
        return None


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class FlagArg(abc.ABC, AttrYamlMixin):
    """
    The data model for the argument or arguments for a flag, for example a flag might have no arguments, it might have
    one argument, it might accept one option from a list of options, or it might accept an arbitrary number of inputs
    """

    def text(self) -> typing.List[str]:
        """
        Returns the text of the argument, e.g. for name generation purposes
        """
        return []

    @abc.abstractmethod
    def get_type(self) -> cli_types.CliType:
        """
        Return a type object indicating the type of data this argument holds. e.g. If it's an array type this will be a
        CliList.
        """
        pass

    @abc.abstractmethod
    def num_args(self) -> int:
        """
        Calculate the multiplicity of this argument
        """
        pass


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class EmptyFlagArg(FlagArg):
    """
    A flag that has no arguments, e.g. `--quiet` that is either present or not present
    """

    def num_args(self) -> int:
        return 0

    def get_type(self):
        return cli_types.CliBoolean()


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class OptionalFlagArg(FlagArg):
    """
    When the flag has multiple arguments, some of which are optional, e.g.
    -I FLOAT[,FLOAT[,INT[,INT]]]
    """

    names: list
    """
    Names of each argument
    """

    separator: str
    """
    Separator between each argument
    """

    def text(self) -> typing.List[str]:
        return list(
            itertools.chain.from_iterable(
                [wordsegment.segment(name) for name in self.names]
            )
        )

    def num_args(self) -> int:
        return len(self.names)

    def get_type(self):
        return cli_types.CliTuple(
            [infer_type(" ".join(wordsegment.segment(arg))) for arg in self.names]
        )


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class SimpleFlagArg(FlagArg):
    """
    When a flag has one single argument, e.g. `-e PATTERN`, where PATTERN is the argument
    """

    name: str
    """
    Name of this argument
    """

    def text(self) -> typing.List[str]:
        return list(wordsegment.segment(self.name))

    def num_args(self) -> int:
        return 1

    def get_type(self):
        return infer_type(" ".join(wordsegment.segment(self.name))) or None


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class RepeatFlagArg(FlagArg):
    """
    When a flag accepts 1 or more arguments, e.g. `--samout SAMOUTS [SAMOUTS ...]`
    """

    name: str
    """
    The name of this argument
    """

    def text(self) -> typing.List[str]:
        return list(wordsegment.segment(self.name))

    def num_args(self) -> int:
        return 1

    def get_type(self):
        t = (
            infer_type(" ".join(wordsegment.segment(self.name)))
            or cli_types.CliString()
        )
        return cli_types.CliList(t)


@yaml_object(yaml)
@attr.s(auto_attribs=True)
class ChoiceFlagArg(FlagArg):
    """
    When a flag accepts one option from a list of options, e.g. `-s {yes,no,reverse}`
    """

    choices: typing.Set[str]
    """
    Set of possible choices that could be used for this argument
    """

    def text(self) -> typing.List[str]:
        return list(
            itertools.chain.from_iterable(
                [wordsegment.segment(name) for name in self.choices]
            )
        )

    def get_type(self):
        e = enum.Enum(
            "".join([choice.capitalize() for choice in self.choices]),
            list(self.choices),
        )
        return cli_types.CliEnum(e)

    def num_args(self) -> int:
        return 1



import abc
import copy
import logging
from typing import Iterable, List, Optional

from pyparsing import ParseBaseException

from aclimatise.execution import Executor
from aclimatise.integration import parse_help
from aclimatise.model import Command

logger = logging.getLogger()


class CliHelpExecutor(Executor):
    """
    This is an abstract class for any executor that works with command-line help conventions like using help flags in
    order to obtain the help text.
    """

    def __init__(
        self,
        flags: Iterable[str] = (["--help"], ["-h"], [], ["--usage"]),
        try_subcommand_flags=True,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.flags = flags
        self.try_subcommand_flags = try_subcommand_flags

    def explore(
        self,
        command: List[str],
        max_depth: int = 2,
        parent: Optional[Command] = None,
    ) -> Optional[Command]:

        logger.info("Exploring {}".format(" ".join(command)))
        best = self.convert(command)
        best.parent = parent

        # Check if this is a valid subcommand
        if parent:
            if best.valid_subcommand():
                logger.info(
                    "{} seems to be a valid subcommand".format(" ".join(command))
                )
            else:
                logger.info(
                    "{} does not seem to be a valid subcommand".format(
                        " ".join(command)
                    )
                )
                return None

        # Recursively call this function on positionals, but only do this if we aren't at max depth
        if best.depth < max_depth:
            # By default we use the best parent help-flag
            child_executor = copy.copy(self)
            child_executor.flags = (
                self.flags if self.try_subcommand_flags else [best.generated_using]
            )

            # Try each *unique* positional
            for positional in {positional.name for positional in best.positional}:
                subcommand = child_executor.explore(
                    command=command + [positional],
                    parent=best,
                    max_depth=max_depth,
                )
                if subcommand is not None:
                    best.subcommands.append(subcommand)
                    # If we had any subcommands then we probably don't have any positionals, or at least don't care about them
                    best.positional = []

        return best

    @abc.abstractmethod
    def execute(self, cmd: List[str]) -> str:
        """
        Executes the provided command and returns a string containing the output
        """
        pass

    def convert(
        self,
        cmd: List[str],
    ) -> Command:
        """
        Determine the best Command instance for a given command line tool, by trying many
        different help flags, such as --help and -h, then return the Command. Use this if you know the command you want to
        parse, but you don't know which flags it responds to with help text. Unlike :py:func:`aclimatise.explore_command`,
        this doesn't even attempt to parse subcommands.

        :param cmd: The command to analyse, e.g. ['wc'] or ['bwa', 'mem']
        :param flags: A list of help flags to try, e.g. ['--help', '-h'], in order how which one you would prefer to use.
        Generally [] aka no flags should be last
        :param executor: A class that provides the means to run a command. You can use the pre-made classes or write your own.
        """
        # For each help flag, run the command and then try to parse it
        logger.info("Trying flags for {}".format(" ".join(cmd)))
        commands = []
        for flag in self.flags:
            help_cmd = cmd + flag
            logger.info("Trying {}".format(" ".join(help_cmd)))
            try:
                final = self.execute(help_cmd)
                result = parse_help(cmd, final, max_length=self.max_length)
                result.generated_using = flag
                commands.append(result)
            except (ParseBaseException, UnicodeDecodeError) as e:
                # If parsing fails, this wasn't the right flag to use
                continue

        # Sort by flags primarily, and if they're equal, return the command with the longest help text, and if they're equal
        # return the command with the most help flags. This helps ensure we get ["bedtools", "--help"] instead of
        # ["bedtools"]
        best = Command.best(commands)
        logger.info(
            "The best help flag seems to be {}".format(
                " ".join(best.command + best.generated_using)
            )
        )
        return best



"""
This module is concerned with running the actual commands so that we can parse their output
"""
import abc
from typing import List, Optional

from aclimatise.model import Command


class Executor(abc.ABC):
    """
    An executor is anything that can take a command such as ["bwa"] or
    ["samtools", "sort"] and return the help output
    """

    def __init__(
        self, timeout: int = 10, raise_on_timout=False, max_length: Optional[int] = 1000
    ):
        """
        :param timeout: Amount of inactivity before the execution will be killed
        :param raise_on_timout: If true, execute will raise a TimeoutError if it
            times out
        """
        # Here we initialise all shared parameters that are used by all executors
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timout
        self.max_length = max_length

    def handle_timeout(self, e: Exception) -> str:
        """
        Subclasses can call this when a timeout has occurred
        :param e: The timeout exception that caused the timeout
        """
        if self.raise_on_timeout:
            raise TimeoutError()
        else:
            return ""

    def explore(
        self,
        command: List[str],
        max_depth: int = 2,
        parent: Optional[Command] = None,
    ) -> Command:
        """
        Given a command to start with, builds a model of this command and all its subcommands (if they exist)
        """
        # If the executor doesn't implement a specific exploration technique, we just execute and ignore subcommands
        return self.convert(command)

    @abc.abstractmethod
    def convert(self, command: List[str]) -> Command:
        """
        Convert a single executable to a Command object, without considering subcommands
        """
        pass



from test.util import skip_not_installed

from aclimatise.execution.man import ManPageExecutor


@skip_not_installed("git")
@skip_not_installed("man")
def test_git():
    cmd = ManPageExecutor(max_length=99999).explore(
        ["git"],
    )
    assert len(cmd.positional) > 20


@skip_not_installed("git")
@skip_not_installed("ls")
def test_ls():
    cmd = ManPageExecutor().explore(
        ["ls"],
    )
    assert {"-A", "--almost-all", "-1", "--context"} <= cmd.all_synonyms



"""
Functions for generating WDL from the CLI data model
"""
import re
from typing import Iterable, List, Set, Tuple

from inflection import camelize
from WDL._grammar import keywords
from wdlgen import (
    ArrayType,
    File,
    Input,
    Output,
    ParameterMeta,
    PrimitiveType,
    Task,
    WdlType,
)

from aclimatise import cli_types, model
from aclimatise.converter import NamedArgument, WrapperGenerator
from aclimatise.model import CliArgument, Command, Flag, Positional
from aclimatise.nlp import wordsegment

#: A regex, borrowed from MiniWDL, that ma
WDL_IDENT = re.compile(r"[a-zA-Z][a-zA-Z0-9_]*")
#: Matches all characters we should remove from a WDL identifier
WDL_STRIP = re.compile(r"(^[^a-zA-Z])|([^a-zA-Z0-9_])")


def escape_wdl_str(text: str):
    """
    Escape literal quotes in a Python string, to become suitable for WDL
    """
    return text.replace('"', '\\"').replace("\n", "\\n")


def flag_to_command_input(
    named_flag: NamedArgument, converter: WrapperGenerator
) -> Task.Command.CommandInput:
    args = dict(name=named_flag.name)

    if isinstance(named_flag.arg, model.Flag):
        args.update(dict(optional=named_flag.arg.optional))
        if isinstance(named_flag.arg.args, model.EmptyFlagArg):
            args.update(dict(true=named_flag.arg.longest_synonym, false=""))
        else:
            args.update(
                dict(
                    prefix=named_flag.arg.longest_synonym,
                )
            )
    elif isinstance(named_flag, model.Positional):
        args.update(dict(optional=False, position=named_flag.position))

    return Task.Command.CommandInput.from_fields(**args)


class WdlGenerator(WrapperGenerator):
    @property
    def suffix(self) -> str:
        return ".wdl"

    case = "snake"

    @property
    def reserved(self) -> Set[Tuple[str, ...]]:
        # Steal the keywords list from miniWDL
        return {tuple(wordsegment.segment(key)) for key in keywords["1.0"]}

    @classmethod
    def format(cls) -> str:
        return "wdl"

    @classmethod
    def type_to_wdl(cls, typ: cli_types.CliType, optional: bool = False) -> WdlType:
        if isinstance(typ, cli_types.CliString):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        elif isinstance(typ, cli_types.CliFloat):
            return WdlType(PrimitiveType(PrimitiveType.kFloat), optional=optional)
        elif isinstance(typ, cli_types.CliBoolean):
            return WdlType(PrimitiveType(PrimitiveType.kBoolean), optional=optional)
        elif isinstance(typ, cli_types.CliInteger):
            return WdlType(PrimitiveType(PrimitiveType.kInt), optional=optional)
        elif isinstance(typ, cli_types.CliFile):
            return WdlType(PrimitiveType(PrimitiveType.kFile), optional=optional)
        elif isinstance(typ, cli_types.CliDir):
            return WdlType(PrimitiveType(PrimitiveType.kDirectory), optional=optional)
        elif isinstance(typ, cli_types.CliTuple):
            if typ.homogenous:
                return WdlType(
                    ArrayType(
                        cls.type_to_wdl(typ.values[0]), requires_multiple=not optional
                    )
                )
            else:
                return WdlType(
                    ArrayType(
                        cls.type_to_wdl(
                            cli_types.CliType.lowest_common_type(typ.values)
                        ),
                        requires_multiple=not optional,
                    )
                )
        elif isinstance(typ, cli_types.CliList):
            return WdlType(
                ArrayType(cls.type_to_wdl(typ.value), requires_multiple=not optional)
            )
        elif isinstance(typ, cli_types.CliEnum):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        else:
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)

    def make_inputs(self, named: Iterable[NamedArgument]) -> List[Input]:
        return [
            Input(
                data_type=self.type_to_wdl(
                    named_arg.arg.get_type(), optional=named_arg.arg.optional
                ),
                name=named_arg.name,
            )
            for named_arg in named
        ]

    def make_command(self, cmd: Command, inputs: List[NamedArgument]) -> Task.Command:
        return Task.Command(
            command=" ".join([WDL_STRIP.sub("_", tok) for tok in cmd.command]),
            inputs=[
                flag_to_command_input(input, self)
                for input in inputs
                if isinstance(input.arg, Positional)
            ],
            arguments=[
                flag_to_command_input(input, self)
                for input in inputs
                if isinstance(input.arg, Flag)
            ],
        )

    def make_parameter_meta(self, named: Iterable[NamedArgument]) -> ParameterMeta:
        params = {}
        for named_arg in named:
            params[named_arg.name] = escape_wdl_str(named_arg.arg.description)

        return ParameterMeta(**params)

    def make_task_name(self, cmd: Command) -> str:
        return camelize(
            "_".join([WDL_STRIP.sub("", token) for token in cmd.command]).replace(
                "-", "_"
            )
        )

    def make_outputs(self, names: List[NamedArgument]) -> List[Output]:
        ret = [
            # We default to always capturing stdout
            Output(data_type=File, name="out_stdout", expression="stdout()")
        ]
        for arg in names:
            typ = arg.arg.get_type()
            if isinstance(typ, cli_types.CliFileSystemType) and typ.output:
                ret.append(
                    Output(
                        data_type=self.type_to_wdl(typ),
                        name="out_" + arg.name,
                        expression='"${{in_{}}}"'.format(arg.name),
                    )
                )

        return ret

    def save_to_string(self, cmd: Command) -> str:
        inputs: List[CliArgument] = [*cmd.named] + (
            [] if self.ignore_positionals else [*cmd.positional]
        )
        names = self.choose_variable_names(inputs)
        runtime = Task.Runtime()
        runtime.add_docker(cmd.docker_image)

        tool = Task(
            name=self.make_task_name(cmd),
            command=self.make_command(cmd, names),
            version="1.0",
            inputs=self.make_inputs(names),
            outputs=self.make_outputs(names),
            parameter_meta=self.make_parameter_meta(names),
            runtime=runtime,
        )

        return tool.get_string()



import typing

from aclimatise.converter import WrapperGenerator
from aclimatise.converter.cwl import CwlGenerator
from aclimatise.converter.janis import JanisGenerator
from aclimatise.converter.wdl import WdlGenerator
from aclimatise.converter.yml import YmlGenerator
from aclimatise.execution import Executor
from aclimatise.execution.docker import DockerExecutor
from aclimatise.execution.local import LocalExecutor
from aclimatise.execution.man import ManPageExecutor
from aclimatise.integration import parse_help
from aclimatise.model import Command, Flag
from deprecated import deprecated

default_executor = LocalExecutor()


@deprecated(
    reason="Please use the explore method on the executors directly. e.g. `LocalExecutor().explore()`"
)
def explore_command(
    cmd: typing.List[str],
    flags: typing.Iterable[str] = (["--help"], ["-h"], [], ["--usage"]),
    parent: typing.Optional[Command] = None,
    max_depth: int = 2,
    try_subcommand_flags=True,
    executor: Executor = default_executor,
) -> typing.Optional[Command]:
    """
    Given a command to start with, builds a model of this command and all its subcommands (if they exist).
    Use this if you know the command you want to parse, you don't know which flags it responds to with help text, and
    you want to include subcommands.
    """
    return executor.explore(cmd, max_depth=max_depth, parent=parent)


__all__ = [
    CwlGenerator,
    WdlGenerator,
    YmlGenerator,
    JanisGenerator,
    LocalExecutor,
    DockerExecutor,
    ManPageExecutor,
    explore_command,
    parse_help,
]



from io import StringIO
from os import PathLike
from pathlib import Path
from typing import Generator, List

import attr

from aclimatise.converter import WrapperGenerator
from aclimatise.model import Command
from aclimatise.yaml import yaml


@attr.s(auto_attribs=True)
class YmlGenerator(WrapperGenerator):
    """
    Internal YML format
    """

    @property
    def suffix(self) -> str:
        return ".yml"

    def save_to_file(self, cmd: Command, path: Path) -> None:
        with path.open("w") as fp:
            yaml.dump(cmd, fp)

    def save_to_string(self, cmd: Command) -> str:
        buffer = StringIO()
        yaml.dump(cmd, buffer)
        return buffer.getvalue()

    @classmethod
    def format(cls) -> str:
        return "yml"
